﻿#region "SoapBox.Core License"
/// <header module="SoapBox.Core"> 
/// Copyright (C) 2009 SoapBox Automation Inc., All Rights Reserved.
/// Contact: SoapBox Automation Licencing (license@soapboxautomation.com)
/// 
/// This file is part of SoapBox Core.
/// 
/// Commercial Usage
/// Licensees holding valid SoapBox Automation Commercial licenses may use  
/// this file in accordance with the SoapBox Automation Commercial License
/// Agreement provided with the Software or, alternatively, in accordance 
/// with the terms contained in a written agreement between you and
/// SoapBox Automation Inc.
/// 
/// GNU Lesser General Public License Usage
/// SoapBox Core is free software: you can redistribute it and/or modify 
/// it under the terms of the GNU Lesser General Public License
/// as published by the Free Software Foundation, either version 3 of the
/// License, or (at your option) any later version.
/// 
/// SoapBox Core is distributed in the hope that it will be useful, 
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU Lesser General Public License for more details.
/// 
/// You should have received a copy of the GNU Lesser General Public License 
/// along with SoapBox Core. If not, see <http://www.gnu.org/licenses/>.
/// </header>
#endregion
        
using System;
using System.Text;
using System.Collections.Generic;
using System.Linq;
using NUnit.Framework; 
using System.ComponentModel;
using System.Collections.ObjectModel;
using System.Windows.Media.Imaging;
using System.Windows.Controls;
using System.Drawing;
using System.Windows.Media;
using Rhino.Mocks;


namespace SoapBox.Core
{
    [TestFixture]
    public class AbstractOptionsItemTest
    {

        #region " Header "
        [Test]
        public void TestHeader()
        {
            //test default value
            ConcreteOptionsItem target = new ConcreteOptionsItem();
            Assert.AreEqual(string.Empty, target.Header);

            //test setting value and property changed event
            bool propertyChanged = false;
            target.PropertyChanged += delegate(object sender, PropertyChangedEventArgs e)
                {
                    if (e.PropertyName == "Header")
                    {
                        Assert.IsFalse(propertyChanged);
                        propertyChanged = true;
                    }
                    else
                    {
                        throw new Exception("Unexpected Property Changed.");
                    }
                };
            string test = Guid.NewGuid().ToString();
            target.SetHeader(test);
            Assert.AreEqual(test, target.Header);
            Assert.IsTrue(propertyChanged);
        }

        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestHeaderNull()
        {
            //should throw exception on null value
            ConcreteOptionsItem target = new ConcreteOptionsItem();
            target.SetHeader(null);
        }
        #endregion

        #region " Items "
        [Test]
        public void TestItems()
        {
            //test default value
            ConcreteOptionsItem target = new ConcreteOptionsItem();
            Assert.IsNotNull(target.Items);
            int i = 0;
            foreach(IOptionsItem a in target.Items)
            {
                i++;
            }
            Assert.AreEqual(0, i);

            //test setting value and property changed event
            bool propertyChanged = false;
            target.PropertyChanged += delegate(object sender, PropertyChangedEventArgs e)
            {
                if (e.PropertyName == "Items")
                {
                    Assert.IsFalse(propertyChanged);
                    propertyChanged = true;
                }
                else
                {
                    throw new Exception("Unexpected Property Changed.");
                }
            };
            var c = new Collection<IOptionsItem>();
            c.Add(MockRepository.GenerateStub<IOptionsItem>());
            target.SetItems(c);
            Assert.AreEqual(c, target.Items);
            Assert.IsTrue(propertyChanged);
        }

        [Test]
        public void Fires_OptionChangedEvent()
        {
            var mocks = new MockRepository();
            var optionsItem1 = mocks.StrictMock<IOptionsItem>();
            optionsItem1.OptionChanged += null;
            LastCall.Constraints(Rhino.Mocks.Constraints.Is.NotNull());
            optionsItem1.OptionChanged -= null;
            LastCall.Constraints(Rhino.Mocks.Constraints.Is.NotNull());
            Expect.Call(delegate { optionsItem1.CancelChanges(); });

            var optionsItem2 = mocks.StrictMock<IOptionsItem>();
            optionsItem2.OptionChanged += null;
            LastCall.Constraints(Rhino.Mocks.Constraints.Is.NotNull());
            Expect.Call(delegate { optionsItem2.CommitChanges(); });

            mocks.ReplayAll();

            ConcreteOptionsItem target = new ConcreteOptionsItem();
            var c = new Collection<IOptionsItem>();
            c.Add(optionsItem1);
            target.SetItems(c);
            Assert.AreEqual(c, target.Items);

            target.CancelChanges();

            c = new Collection<IOptionsItem>();
            c.Add(optionsItem2);
            target.SetItems(c);
            Assert.AreEqual(c, target.Items);

            int eventRaisedCount = 0;
            target.OptionChanged += new EventHandler((s, e) =>
            {
                eventRaisedCount++;
            });
            optionsItem2.Raise(x => x.OptionChanged += null, optionsItem2, EventArgs.Empty);
            Assert.AreEqual(1, eventRaisedCount);

            target.CommitChanges();

            mocks.VerifyAll();
        }

        #endregion

        #region " Pad "
        [Test]
        public void TestPad()
        {
            //test default value
            ConcreteOptionsItem target = new ConcreteOptionsItem();
            Assert.IsNull(target.Pad);

            //test setting value and property changed event
            bool propertyChanged = false;
            target.PropertyChanged += delegate(object sender, PropertyChangedEventArgs e)
            {
                if (e.PropertyName == "Pad")
                {
                    Assert.IsFalse(propertyChanged);
                    propertyChanged = true;
                }
                else
                {
                    throw new Exception("Unexpected Property Changed.");
                }
            };
            IOptionsPad p = new ConcreteOptionsPad();
            target.SetPad(p);
            Assert.AreEqual(p, target.Pad);
            Assert.IsTrue(propertyChanged);
        }

        #endregion

        #region " CommitChanges "
        [Test]
        public void TestCommitChanges()
        {
            ConcreteOptionsItem target = new ConcreteOptionsItem();
            target.CommitChanges();
        }
        #endregion

        #region " CancelChanges "
        [Test]
        public void TestCancelChanges()
        {
            ConcreteOptionsItem target = new ConcreteOptionsItem();
            target.CancelChanges();
        }
        #endregion
    }
}
